import asyncio

from sus import SusServer
from sus.common.util import logger_config


def server_main():
    logger_config()
    server = SusServer(("0.0.0.0", args.port), open(args.private_key, "r").read().strip())
    asyncio.run(server.start())


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description="SUS")
    command = parser.add_subparsers(dest="command")

    server = command.add_parser("server", help="Run SUS server")
    server.add_argument("-k", "--private-key", help="Private key file", required=True)
    server.add_argument("-p", "--port", help="Server port", default=42069, type=int)

    client = command.add_parser("client", help="Run SUS client")
    client.add_argument("-k", "--public-key", help="Public key", required=True)
    client.add_argument("-s", "--server", help="Server address", default="localhost")
    client.add_argument("-p", "--port", help="Server port", default=42069, type=int)
    client.add_argument("-a", "--application-protocol", help="Application protocol", default=b"cliq")

    args = parser.parse_args()

    if args.command == "client":
        from sus.client import SusClient

        logger_config()
        client = SusClient((args.server, int(args.port)), open(args.public_key, "r").read().strip(), b"fuckyou")
        asyncio.run(client.start())
        client.send(b"fuckyou" * 3000)
    else:
        server_main()



"""
The client protocol is responsible for handling the connection to the server.
It encrypts and decrypts messages, and handles the handshake. It also handles
splitting messages into packets and reassembling them.
"""

import asyncio
import logging
from typing import Iterable, Optional

from cryptography.hazmat.primitives import poly1305
from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.ciphers.algorithms import ChaCha20

from sus.common.exceptions import MalformedPacket
from sus.common.globals import CLIENT_ENC_NONCE, CLIENT_MAC_NONCE, SERVER_ENC_NONCE, SERVER_MAC_NONCE
from sus.common.util import ConnectionState, MessageHandler, Wallet, now, trail_off


class SusClientProtocol(asyncio.DatagramProtocol):
    """
    This class is responsible for handling the UDP protocol.
    """
    transport: asyncio.DatagramTransport
    state: ConnectionState

    def __init__(self, wallet: Wallet, protcol_id: bytes,
                 handlers: Optional[Iterable[MessageHandler]] = None):
        """
        Initializes the client protocol.
        :param wallet: wallet containing the client's keys
        :param protcol_id: protocol ID (any bytestring)
        :param handlers: message handlers, called when a message is received
        """
        super().__init__()

        self.wallet = wallet
        self.protocol_id = protcol_id
        self.state = ConnectionState.INITIAL

        self.logger = logging.getLogger(f"sus-cl")

        self.last_seen = now()

        self.cl_enc = Cipher(ChaCha20(wallet.shared_secret, b"\x00" * 8 + CLIENT_ENC_NONCE), None).encryptor()
        self.sr_enc = Cipher(ChaCha20(wallet.shared_secret, b"\x00" * 8 + SERVER_ENC_NONCE), None).decryptor()
        self.cl_mac = Cipher(ChaCha20(wallet.shared_secret, b"\x00" * 8 + CLIENT_MAC_NONCE), None).encryptor()
        self.sr_mac = Cipher(ChaCha20(wallet.shared_secret, b"\x00" * 8 + SERVER_MAC_NONCE), None).decryptor()

        self.client_message_id = 0
        self.server_message_id = 0
        self.client_packet_id = 0
        self.server_packet_id = 0

        self.mtu_estimate = 1500

        self.message_handlers: set[MessageHandler] = set(handlers or [])
        self.handshake_event = asyncio.Event()
        self.diconnection_event = asyncio.Event()

    def add_message_handler(self, handler: MessageHandler):
        self.message_handlers.add(handler)

    def connection_made(self, transport: asyncio.DatagramTransport):
        """
        This function is called when the connection is established.
        :param transport: transport object, used to send and receive packets
        """
        self.transport = transport
        self.state = ConnectionState.HANDSHAKE
        self.send(self.protocol_id, self.wallet.token)
        self.state = ConnectionState.CONNECTED
        self.last_seen = now()
        self.logger.debug("Handshake complete")
        self.handshake_event.set()

    def datagram_received(self, data: bytes, _addr: tuple[str, int]):
        """
        This function is called when a packet is received.
        :param data: packet data
        :param _addr: originating address (always the server, unused)
        """

        if self.state == ConnectionState.CONNECTED:
            pid, message = self.__verify_and_decrypt(data)
            self.logger.info(f">>> {trail_off(message.decode('utf-8')) if message else None}")
            self.handle_message(pid, message)

    def __verify_and_decrypt(self, data: bytes) -> tuple[None, None] | tuple[int, bytes]:
        """
        This function is responsible for verifying the packet and decrypting it.
        :param data: data to verify and decrypt
        :return: packet ID and message, or None if the packet is invalid
        """
        try:
            key = self.sr_mac.update(b"\x00" * 32)
            p_id = data[:8]
            payload = data[8:-16]
            tag = data[-16:]
            frame = p_id + payload
            poly1305.Poly1305.verify_tag(key, frame, tag)
        except poly1305.InvalidSignature:
            self.logger.error("Invalid signature")
            return None, None

        p_id = int.from_bytes(p_id, "little")
        message_bytes = self.sr_enc.update(payload)
        message_length = int.from_bytes(message_bytes[:4], "little")
        message: bytes = message_bytes[4:message_length + 4]
        # self.logger.info(f"Received message {p_id} ({message_length} bytes)")
        self.last_seen = now()
        self.server_packet_id = p_id
        return p_id, message

    def __split_message(self, data: bytes) -> list[bytes]:
        """
        This function is responsible for splitting a message into packets.
        :param data: data to split
        :return: list of packets
        """
        packet_length = self.mtu_estimate - 24
        return [data[i:i + packet_length] for i in range(0, len(data), packet_length)]

    def __encrypt_and_tag(self, data: bytes, token: bytes) -> list[bytes]:
        """
        This function is responsible for encrypting and tagging a message. Uses the ChaCha20-Poly1305 AEAD to
        encrypt and authenticate the message.
        :param data: data to encrypt
        :param token: optional token to include in the first packet
        :return: packets containing the encrypted and tagged message to send to the server
        """
        message_bytes = len(data).to_bytes(4, "little") + data
        packet_length = self.mtu_estimate - 24
        padded_message_bytes = message_bytes  # + b"\x00" * (
        # packet_length - ((len(message_bytes) + len(token)) % packet_length))

        ciphertext = self.cl_enc.update(padded_message_bytes)
        self.logger.debug(f"--- {trail_off(ciphertext.hex())}")
        if token:
            self.logger.debug(f"TOK {trail_off(token.hex())}")

        payloads = self.__split_message(token + ciphertext)

        packets = []
        for payload in payloads:
            key = self.cl_mac.update(b"\x00" * 32)
            p_id = self.client_packet_id.to_bytes(8, "little")
            frame = p_id + payload
            tag = poly1305.Poly1305.generate_tag(key, frame)
            packets.append(frame + tag)
            self.client_packet_id += 1
        return packets

    def send(self, data: bytes, token: bytes = b""):
        """
        This function is responsible for sending a message to the server.
        :param data: data to send
        :param token: token to include in the first packet. DO NOT INCLUDE THE TOKEN IN SUBSEQUENT PACKETS --
                      a MalformedPacket error will be raised!
        :raises MalformedPacket: if the token is included in a subsequent packet
        """
        if self.state not in (ConnectionState.CONNECTED, ConnectionState.HANDSHAKE):
            return
        if token and self.client_packet_id != 0:
            raise MalformedPacket("Token can only be included in the first packet")
        self.logger.info(f"<<< {trail_off(data.decode('utf-8'))}")
        packets = self.__encrypt_and_tag(data, token)
        self.logger.info(f"Sending {len(data)} bytes in {len(packets)} packets")
        for packet in packets:
            self.transport.sendto(packet, None)

    def disconnect(self):
        """
        Disconnects the client from the server.
        """
        self.logger.warning("Disconnecting from server...")
        self.transport.close()

    def connection_lost(self, exc):
        """
        Called when the connection is lost. Sets the disconnection event.
        :param exc: exception raised, if any
        """
        self.logger.warning("Connection to server lost")
        if exc:
            self.logger.exception(exc)
            self.state = ConnectionState.ERROR
        else:
            self.state = ConnectionState.DISCONNECTED
        self.diconnection_event.set()

    def handle_message(self, pid: int, message: bytes):
        """
        Calls all message handlers asynchronously.
        :param pid: packet ID
        :param message: message bytes
        :return:
        """
        asyncio.gather(*[handler(("", 0), pid, message) for handler in self.message_handlers])



"""
This file contains global variables used by SUS.
"""

CLIENT_ENC_NONCE = b"cliqencc"
"Client encryption nonce"

SERVER_ENC_NONCE = b"cliqencs"
"Server encryption nonce"

CLIENT_MAC_NONCE = b"cliqmacc"
"Client MAC nonce"

SERVER_MAC_NONCE = b"cliqmacs"
"Server MAC nonce"



"""
The client class.
"""

import asyncio
import logging
import socket
from os import urandom
from typing import Iterable

from blake3 import blake3
from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PrivateKey, X25519PublicKey
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat

from sus.client.protocol import SusClientProtocol
from sus.common.exceptions import MalformedPacket
from sus.common.util import ConnectionState, MessageHandler, Wallet


class SusClient:
    """
    This class is responsible for managing the client.
    """
    protocol: SusClientProtocol

    def __init__(self, addr: tuple[str, int], ppks: str, protocol_id: bytes):
        """
        Initializes the client.
        :param addr: Server address
        :param ppks: Server public key
        :param protocol_id:  Protocol ID (any bytestring)
        """
        self.server_addr = addr
        self.ppks = X25519PublicKey.from_public_bytes(bytes.fromhex(ppks))
        self.protocol_id = protocol_id

        self.logger = logging.getLogger(f"sussus")

    def __del__(self):
        self.disconnect()

    @property
    def connected(self):
        """
        True if the client is connected to the server.
        """
        return hasattr(self, "protocol") and self.protocol.state == ConnectionState.CONNECTED

    async def start(self, handlers: Iterable[MessageHandler] = None):
        """
        This coroutine is responsible for starting the client. Blocks until the client is connected.
        It also registers message handlers, called when a message is received.
        :param handlers:
        :return:
        """
        await self.connect()
        for handler in handlers or []:
            self.protocol.add_message_handler(handler)

    def __key_exchange(self, epks_ns_port: bytes, wallet: Wallet) -> Wallet:
        """
        This function is responsible for performing the key exchange.
        :param epks_ns_port: received (epks, ns, port) from server
        :param wallet: wallet containing the client's keys
        :return: wallet containing the shared secret
        """

        if len(epks_ns_port) != 40:
            raise MalformedPacket("Invalid key response length")
        # 4. receive (epks, ns, port) from server
        wallet.epks = X25519PublicKey.from_public_bytes(epks_ns_port[:32])
        wallet.ns = epks_ns_port[32:40]
        self.logger.info("received keys, starting handshake")
        # 5. compute ecps = X25519(eskc, ppks)
        ecps = wallet.eskc.exchange(wallet.ppks)
        eces = wallet.eskc.exchange(wallet.epks)
        # 6. compute key = H(eces, ecps, nc, ns, ppks, epks, epkc)
        wallet.shared_secret = blake3(
            eces + ecps + wallet.nc + wallet.ns +
            wallet.ppks.public_bytes(Encoding.Raw, PublicFormat.Raw) +
            wallet.epks.public_bytes(Encoding.Raw, PublicFormat.Raw) +
            wallet.epkc.public_bytes(Encoding.Raw, PublicFormat.Raw)).digest()
        self.logger.info("shared secret: %s", wallet.shared_secret.hex())

        # 7. compute token = H(epkc, epks, nc, ns)
        self.logger.info("\n".join([
            f"epkc: {wallet.epkc.public_bytes(Encoding.Raw, PublicFormat.Raw).hex()}",
            f"epks: {wallet.epks.public_bytes(Encoding.Raw, PublicFormat.Raw).hex()}",
            f"nc: {wallet.nc.hex()}",
            f"ns: {wallet.ns.hex()}"
        ]))
        wallet.token = blake3(wallet.epkc.public_bytes(Encoding.Raw, PublicFormat.Raw) +
                              wallet.epks.public_bytes(Encoding.Raw, PublicFormat.Raw) +
                              wallet.nc + wallet.ns).digest()
        return wallet

    async def connect(self):
        """
        This coroutine is responsible for connecting to the server.
        Performs the key exchange and starts the handshake.
        """
        self.logger.info(f"connecting to server ({self.server_addr[0]}:{self.server_addr[1]})")

        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.connect(self.server_addr)
        sock.setblocking(False)
        sock.settimeout(5)

        eskc = X25519PrivateKey.generate()
        epkc = eskc.public_key()
        nc = urandom(8)
        wallet = Wallet(ppks=self.ppks, eskc=eskc, epkc=epkc, nc=nc)

        # try:
        sock.send(wallet.epkc.public_bytes(Encoding.Raw, PublicFormat.Raw) + wallet.nc)
        data = sock.recv(40)
        # except (ConnectionError, TimeoutError):
        #     self.logger.error("failed to connect to server")
        # return False

        wallet = self.__key_exchange(data, wallet)

        self.logger.info("received keys, starting handshake")

        _, self.protocol = await asyncio.get_event_loop().create_datagram_endpoint(
            lambda: SusClientProtocol(wallet, self.protocol_id),
            sock=sock
        )
        await self.protocol.handshake_event.wait()
        # return True

    def send(self, data: bytes):
        """
        Sends a message to the server.
        :param data: message to send as bytes
        """
        if not self.protocol:
            self.logger.warning("not connected to server")
            return
        self.protocol.send(data)

    def disconnect(self):
        """
        Disconnects from the server.
        """
        if not hasattr(self, "protocol"):
            self.logger.warning("not connected to server")
            return
        try:
            asyncio.get_running_loop()
            self.protocol.disconnect()
        except RuntimeError:  # not running in event loop
            pass
        self.logger.info(f"disconnected from server ({self.server_addr[0]}:{self.server_addr[1]})")

    async def keep_alive(self):
        """
        Convenience coroutine that waits until the client is disconnected.
        """
        if not hasattr(self, "protocol"):
            self.logger.warning("not connected to server")
            return
        try:
            await self.protocol.diconnection_event.wait()
        except asyncio.CancelledError:
            self.logger.info("exiting...")



"""
This module contains the OnePortProtocol class.
"""
import asyncio
import logging
from typing import Iterable

from sus.common.exceptions import HandsakeError, MalformedPacket
from sus.common.util import Address, ConnectionID, MessageHandler, Wallet
from sus.server.handler import ClientHandler


class OnePortProtocol(asyncio.DatagramProtocol):
    """
    This class is responsible for handling the UDP protocol.
    It matches incoming packets to clients and handles the handshake.
    """
    __transport: asyncio.DatagramTransport

    def __init__(self, wallet: Wallet, message_handlers: Iterable[MessageHandler]):
        super().__init__()
        self.__wallet = wallet
        self.__message_handlers = message_handlers

        self.__addr_to_conn_id: dict[Address, ConnectionID] = dict()
        self.__clients: dict[ConnectionID, ClientHandler] = dict()
        self.__logger = logging.getLogger(f"oneportsus")

        self.closed = asyncio.Event()

    def connection_made(self, transport: asyncio.DatagramTransport):
        self.__transport = transport
        self.__logger.info(f"Listening on port {transport.get_extra_info('sockname')[1]}")

    def error_received(self, exc):
        self.__logger.exception(exc)

    def datagram_received(self, data, addr):
        if addr not in self.__addr_to_conn_id:
            try:
                c = ClientHandler(addr, self.__transport, self.__wallet, self.__message_handlers)
                c.handle(data)
            except (HandsakeError, MalformedPacket):
                self.__logger.error(f"Handshake failed with {addr}")
                return
            self.__addr_to_conn_id[addr] = c.connection_id
            self.__clients[c.connection_id] = c
            return

        connection_id = self.__addr_to_conn_id[addr]
        handler = self.__clients[connection_id]

        try:
            handler.handle(data)
        except MalformedPacket:
            self.__logger.error(f"Malformed packet from {addr}")
            del self.__clients[connection_id]

    def send(self, data: bytes, addr: tuple[str, int]):
        if addr not in self.__clients:
            self.__logger.error(f"Attempted to send to {addr} but they are not connected")
            return
        self.__clients[self.__addr_to_conn_id[addr]].send(data)

    def add_message_handler(self, handler: MessageHandler, addr: tuple[str, int]):
        self.__clients[self.__addr_to_conn_id[addr]].add_message_handler(handler)

    def clean(self):
        """
        Removes inactive clients.
        """
        for addr, conn_id in list(self.__addr_to_conn_id.items()):
            client = self.__clients.get(conn_id, None)
            if client is None:
                del self.__addr_to_conn_id[addr]
                continue
            if not client.is_alive:
                self.__logger.warning(f"Removing inactive client {addr}")
                del self.__addr_to_conn_id[addr]
                self.__clients[conn_id].disconnect()
                del self.__clients[conn_id]

    def close(self):
        self.__transport.close()

    def connection_lost(self, exc):
        self.__logger.warning("Connection closed")
        self.closed.set()



"""
This module contains all the exceptions used in the sus package.
"""


class MalformedKeyRequest(Exception):
    """Raised when a key request is malformed"""
    pass


class MalformedPacket(Exception):
    """Raised when a packet is malformed"""
    pass


class HandsakeError(Exception):
    """Raised when a handshake fails"""
    pass



"""
SUS server implementation.
"""

import asyncio
import logging
import os.path
from typing import Iterable

from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PrivateKey
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat

from sus.common.util import Address, MessageHandler, Wallet
from sus.server.protocol import OnePortProtocol

GARABAGE_COLLECTOR_INTERVAL = 15  # seconds


class SusServer:
    """
    This class is responsible for managing the server.
    """
    __protocol: OnePortProtocol

    def __init__(self, addr: Address, psks: str):
        """
        Initializes the server.
        :param addr: Tuple containing the address and port to listen on
        :param psks: Hex encoded private key
        """
        self.__addr = addr
        self.__logger = logging.getLogger("gatekeeper")

        if not psks or len(psks) != 64:
            self.__logger.warning("Invalid or empty private key provided. Generating a new one.")
            self.__psks = X25519PrivateKey.generate()
        else:
            self.__psks = X25519PrivateKey.from_private_bytes(bytes.fromhex(psks))
        self.__ppks = self.__psks.public_key()

        if not os.path.exists("server.pub"):
            self.__logger.warning("Server key not found, creating a new one")

            # presumably, this is the first time the server is run.
            # print a helpful message to the user.
            print("\nSus\n===\n")
            print("Welcome to SUS!")
            print("This is the first time you're running the server. A new keypair will be generated.")
            print("Please copy the following public key to the client:")
            print(self.__ppks.public_bytes(Encoding.Raw, PublicFormat.Raw).hex())
            print("")

        with open("server.pub", "w") as f:
            f.write(self.__ppks.public_bytes(Encoding.Raw, PublicFormat.Raw).hex())

    @property
    def public_key(self):
        """Server public key"""
        return self.__ppks.public_bytes(Encoding.Raw, PublicFormat.Raw).hex()

    @property
    def address(self):
        """Server address"""
        return self.__addr

    async def __garbage_collector(self):
        """
        This coroutine is responsible for cleaning up inactive clients.
        """
        try:
            while not self.__protocol.closed.is_set():
                await asyncio.sleep(GARABAGE_COLLECTOR_INTERVAL)
                self.__protocol.clean()
        except asyncio.CancelledError:
            pass

    async def start(self, message_handlers: Iterable[MessageHandler] = None):
        """
        This coroutine is responsible for starting the server.
        :param message_handlers: An iterable of message handlers, which are called when a message is received.
        """
        self.__logger.info("Starting server")
        self.__logger.info(f"public key: {self.__ppks.public_bytes(Encoding.Raw, PublicFormat.Raw).hex()}")

        wallet = Wallet(ppks=self.__ppks, psks=self.__psks)

        # create a protocol instance, this will handle all incoming packets
        _, self.__protocol = await asyncio.get_running_loop().create_datagram_endpoint(
            lambda: OnePortProtocol(wallet, message_handlers if message_handlers else []),
            self.__addr)

        # start the garbage collector.
        gc_task = asyncio.create_task(self.__garbage_collector())

        # if message handlers were emoty, warn the user.
        if not message_handlers:
            self.__logger.warning("No message handlers were provided. You will not receive any messages from clients.")
            self.__logger.warning("Please specify message handlers when calling start().")
        # we're done here, wait for the protocol to close.
        try:
            await self.__protocol.closed.wait()
        except asyncio.CancelledError:
            self.stop()
        finally:
            gc_task.cancel()

    def send(self, addr: Address, msg: bytes):
        """
        Schedules a message to be sent to a client.
        :param addr: Client address
        :param msg: Message to send
        """
        self.__protocol.send(msg, addr)

    def stop(self):
        """
        Stops the server.
        """
        self.__logger.warning("Shutting down")
        self.__protocol.close()



"""
Utility functions and classes.
"""

from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from logging import WARNING, basicConfig, getLogger
from typing import Any, Awaitable, Callable, Optional, Type, TypeAlias

from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PrivateKey, X25519PublicKey

Address = tuple[str, int]
MessageHandler: Type = Callable[[Address, int, bytes], Awaitable[Any]]


def trail_off(msg: str, length: int = 40):
    """
    Truncates a string to a given length and adds an ellipsis.
    :param msg: message to truncate
    :param length: maximum length of the string (including ellipsis)
    :return: truncated string
    """
    if len(msg) > length:
        msg = msg[:length - 3] + "..."
    return msg


def logger_config():
    """Default logger configuration."""
    basicConfig(level=WARNING, format="%(asctime)s | %(name)s - %(levelname)8s : %(message)s")
    getLogger('asyncio').setLevel(WARNING)


def now():
    """Returns the current timestamp."""
    return datetime.now().timestamp()


@dataclass
class Wallet:
    """
    Convenience class for storing keys and other data.
    """

    ppks: X25519PublicKey
    """The server's public key."""
    psks: Optional[X25519PrivateKey] = None
    """The server's private key."""

    esks: Optional[X25519PrivateKey] = None
    """ephemeral server private key"""
    epks: Optional[X25519PublicKey] = None
    """ephemeral server public key"""
    ns: Optional[bytes] = None
    """server nonce"""

    eskc: Optional[X25519PrivateKey] = None
    """ephemeral client private key"""
    epkc: Optional[X25519PublicKey] = None
    """ephemeral client public key"""
    nc: Optional[bytes] = None
    """client nonce"""

    token: Optional[bytes] = None
    """Verification token"""
    shared_secret: Optional[bytes] = None
    """Shared secret"""


ConnectionID: TypeAlias = int


class ConnectionState(Enum):
    """
    Enum representing the state of the connection.
    """
    ERROR = -1
    INITIAL = 0
    HANDSHAKE = 1
    CONNECTED = 2
    DISCONNECTED = 3



import asyncio
import collections
import logging
from os import urandom
from typing import Iterable

from blake3 import blake3
from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives import poly1305
from cryptography.hazmat.primitives.asymmetric.x25519 import (
    X25519PrivateKey,
    X25519PublicKey,
)
from cryptography.hazmat.primitives.ciphers import (
    AEADDecryptionContext,
    AEADEncryptionContext,
    Cipher,
)
from cryptography.hazmat.primitives.ciphers.algorithms import ChaCha20
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat

from sus.common.exceptions import HandsakeError, MalformedPacket
from sus.common.globals import (
    CLIENT_ENC_NONCE,
    CLIENT_MAC_NONCE,
    SERVER_ENC_NONCE,
    SERVER_MAC_NONCE,
)
from sus.common.util import (
    ConnectionID,
    ConnectionState,
    MessageHandler,
    Wallet,
    now,
    trail_off,
)
from sus import PYTHON_VERSION


class ClientHandler:
    """
    This class is responsible for handling clients.
    One instance of this class is created for each client.
    """

    __cl_enc: AEADDecryptionContext
    __sr_enc: AEADEncryptionContext
    __cl_mac: AEADDecryptionContext
    __sr_mac: AEADEncryptionContext

    __protocol: bytes
    __connection_id: ConnectionID

    def __init__(
        self,
        addr: tuple[str, int],
        transport: asyncio.DatagramTransport,
        wallet: Wallet,
        message_handlers: Iterable[MessageHandler],
        max_packets: int = 8,
    ):
        self.__last_seen = now()
        self.__addr = addr
        self.__transport = transport
        self.__message_handlers = set(message_handlers)
        self.__state = ConnectionState.INITIAL

        self.__logger = logging.getLogger(f"{addr[0]}:{addr[1]}")

        self.__logger.info(f"New client {addr}")

        wallet.esks = X25519PrivateKey.generate()
        wallet.epks = wallet.esks.public_key()
        wallet.ns = urandom(8)

        self.__wallet = wallet

        self.__client_message_id = 0
        self.__server_message_id = 0
        self.__incoming_packet_id = 0
        self.__outgoing_packet_id = 0

        self.__mtu_estimate = 1500

        self.__incoming_buffer: collections.deque[bytes] = collections.deque(
            maxlen=max_packets
        )
        self.__incoming_keys: collections.deque[bytes] = collections.deque(
            maxlen=max_packets
        )
        self.__pending_message_length = 0
        self.__pending_message_buffer = bytearray()
        self.__outgoing_buffer: bytearray = bytearray()
        self.__send_loop_task: asyncio.Task = asyncio.create_task(self.__send_loop())

    def __del__(self):
        self.__logger.info(f"Client {self.__addr} disconnected")
        self.__send_loop_task.cancel()

    @property
    def is_alive(self):
        """
        True if the client is not in an error state and has been seen in the last 5 seconds.
        """
        return (
            self.__state not in (ConnectionState.ERROR, ConnectionState.DISCONNECTED)
            and now() - self.__last_seen < 5
        )

    @property
    def addr(self):
        """Client address."""
        return self.__addr

    @property
    def last_seen(self):
        """Last time the client was seen, in seconds since the epoch."""
        return self.__last_seen

    @property
    def protocol(self):
        return self.__protocol

    @property
    def max_packets(self):
        return self.__incoming_buffer.maxlen

    @property
    def connection_id(self):
        return self.__connection_id

    async def __send_loop(self):
        try:
            while True:
                await asyncio.sleep(1)
                self.__flush()
        except asyncio.CancelledError:
            return

    def __gen_connection_id(self, channel_id: int = 0) -> ConnectionID:
        wallet = self.__wallet

        # figure out a way to determinisitically generate connection_id
        # for now, this just hashes the shared secret with the channel ID
        return int.from_bytes(
            blake3(
                wallet.epkc.public_bytes(Encoding.Raw, PublicFormat.Raw)
                + wallet.epkc.public_bytes(Encoding.Raw, PublicFormat.Raw)
                + int.to_bytes(channel_id)
            ).digest()[:8],
            "little",
            signed=False,
        )

    def __key_exchange(self):
        wallet = self.__wallet
        eces = wallet.esks.exchange(wallet.epkc)
        ecps = wallet.psks.exchange(wallet.epkc)
        wallet.shared_secret = blake3(
            eces
            + ecps
            + wallet.nc
            + wallet.ns
            + wallet.ppks.public_bytes(Encoding.Raw, PublicFormat.Raw)
            + wallet.epks.public_bytes(Encoding.Raw, PublicFormat.Raw)
            + wallet.epkc.public_bytes(Encoding.Raw, PublicFormat.Raw)
        ).digest()
        self.__logger.debug(f"shared_secret: {wallet.shared_secret.hex()}")
        # noinspection DuplicatedCode
        self.__cl_enc = Cipher(
            ChaCha20(wallet.shared_secret, b"\x00" * 8 + CLIENT_ENC_NONCE), None
        ).decryptor()
        self.__sr_enc = Cipher(
            ChaCha20(wallet.shared_secret, b"\x00" * 8 + SERVER_ENC_NONCE), None
        ).encryptor()
        self.__cl_mac = Cipher(
            ChaCha20(wallet.shared_secret, b"\x00" * 8 + CLIENT_MAC_NONCE), None
        ).decryptor()
        self.__sr_mac = Cipher(
            ChaCha20(wallet.shared_secret, b"\x00" * 8 + SERVER_MAC_NONCE), None
        ).encryptor()

        for i in range(self.max_packets):
            self.__incoming_keys.append(self.__cl_mac.update(b"\x00" * 32))

    def __reform_messages(self, data: bytes) -> [bytes]:
        buffer = self.__pending_message_buffer + data
        pending_length = self.__pending_message_length
        messages = []
        # self.__logger.debug(f"buffer: {buffer.hex()}")
        while len(buffer) >= pending_length:
            if pending_length == 0:
                if len(buffer) < 4:
                    break
                pending_length = int.from_bytes(buffer[:4], "little")
                buffer = buffer[4:]
            else:
                messages.append(bytes(buffer[:pending_length]))
                buffer = buffer[pending_length:]
                pending_length = 0
        self.__pending_message_buffer = buffer
        self.__pending_message_length = pending_length
        # self.__logger.debug(f"pending: {pending_length}")
        return messages

    def __verify_and_decrypt(self, data: bytes) -> bytes | None:
        try:
            p_id = int.from_bytes(data[:8], "little")
            if (
                p_id > self.__incoming_packet_id + self.max_packets
                or p_id < self.__incoming_packet_id
            ):
                self.__logger.error("Packet {p_id} is too far ahead")
                return None
            key = self.__incoming_keys[p_id - self.__incoming_packet_id]
            payload = data[8:-16]
            tag = data[-16:]
            poly1305.Poly1305.verify_tag(key, data[:8] + payload, tag)
        except InvalidSignature:
            self.__logger.error("Invalid signature")
            return None

        if PYTHON_VERSION[1] >= 10:
            match p_id:
                case 0:
                    # special case for first packet
                    payload = payload[32:]
                    self.__logger.debug(f"--- {trail_off(payload.hex())}")
                    self.__incoming_packet_id = 1
                    message_bytes = self.__cl_enc.update(payload)

                    self.__incoming_keys.popleft()
                    self.__incoming_keys.append(self.__cl_mac.update(b"\x00" * 32))
                    return message_bytes
                case self.__incoming_packet_id:
                    self.__logger.debug(f"--- {trail_off(payload.hex())}")

                    # packet has already been verified
                    _ = self.__incoming_keys.popleft()
                    self.__incoming_keys.append(self.__cl_mac.update(b"\x00" * 32))
                    self.__incoming_packet_id += 1
                    message_bytes = self.__cl_enc.update(payload)
                    buffer = bytearray(message_bytes)

                    while (
                        self.__incoming_buffer and self.__incoming_buffer[0] is not None
                    ):
                        packet = self.__incoming_buffer.popleft()
                        _ = self.__incoming_keys.popleft()
                        assert (
                            int.from_bytes(packet[:8], "little")
                            == self.__incoming_packet_id
                        )
                        self.__incoming_packet_id += 1
                        payload = packet[8:-16]
                        message_bytes = self.__cl_enc.update(payload)
                        buffer.extend(message_bytes)

                    return bytes(buffer)

                case _:
                    self.__incoming_keys.insert(
                        p_id - self.__incoming_packet_id, self.__cl_mac.update(payload)
                    )
                    self.__incoming_buffer.insert(
                        p_id - self.__incoming_packet_id, data
                    )
                    return None
        else:
            if p_id == 0:
                # special case for first packet
                payload = payload[32:]
                self.__logger.debug(f"--- {trail_off(payload.hex())}")
                self.__incoming_packet_id = 1
                message_bytes = self.__cl_enc.update(payload)

                self.__incoming_keys.popleft()
                self.__incoming_keys.append(self.__cl_mac.update(b"\x00" * 32))
                return message_bytes
            elif p_id == self.__incoming_packet_id:
                self.__logger.debug(f"--- {trail_off(payload.hex())}")

                # packet has already been verified
                _ = self.__incoming_keys.popleft()
                self.__incoming_keys.append(self.__cl_mac.update(b"\x00" * 32))
                self.__incoming_packet_id += 1
                message_bytes = self.__cl_enc.update(payload)
                buffer = bytearray(message_bytes)

                while self.__incoming_buffer and self.__incoming_buffer[0] is not None:
                    packet = self.__incoming_buffer.popleft()
                    _ = self.__incoming_keys.popleft()
                    assert (
                        int.from_bytes(packet[:8], "little")
                        == self.__incoming_packet_id
                    )
                    self.__incoming_packet_id += 1
                    payload = packet[8:-16]
                    message_bytes = self.__cl_enc.update(payload)
                    buffer.extend(message_bytes)

                return bytes(buffer)
            else:
                self.__incoming_keys.insert(
                    p_id - self.__incoming_packet_id, self.__cl_mac.update(payload)
                )
                self.__incoming_buffer.insert(p_id - self.__incoming_packet_id, data)
                return None

    def __encrypt_and_tag(self, data: bytes) -> list[bytes]:
        message_bytes = len(data).to_bytes(4, "little") + data
        packet_length = self.__mtu_estimate - 24
        padded_message_bytes = message_bytes  # + b"\x00" * (
        # packet_length - ((len(message_bytes)) % packet_length))

        ciphertext = self.__sr_enc.update(padded_message_bytes)
        self.__logger.debug(f"--- {trail_off(ciphertext.hex())}")

        payloads = self.__split_message(ciphertext)

        packets = []
        for payload in payloads:
            key = self.__sr_mac.update(b"\x00" * 32)
            p_id = self.__incoming_packet_id.to_bytes(8, "little")
            frame = p_id + payload
            tag = poly1305.Poly1305.generate_tag(key, frame)
            packets.append(frame + tag)
            self.__incoming_packet_id += 1
        return packets

    def __split_message(self, data: bytes) -> list[bytes]:
        packet_length = self.__mtu_estimate - 24
        return [data[i : i + packet_length] for i in range(0, len(data), packet_length)]

    def __initial(self, data):
        if len(data) != 40:
            self.__logger.error("Invalid handshake packet")
            raise MalformedPacket("Invalid handshake packet")
        wallet = self.__wallet
        wallet.epkc = X25519PublicKey.from_public_bytes(data[:32])
        wallet.nc = data[32:]

        self.__connection_id = self.__gen_connection_id()
        self.__state = ConnectionState.HANDSHAKE
        self.__transport.sendto(
            (wallet.epks.public_bytes(Encoding.Raw, PublicFormat.Raw) + wallet.ns),
            self.__addr,
        )
        self.__logger.info(f"connection ID: {self.__connection_id}")

    def __handshake(self, data) -> None:
        if len(data) < 40:
            self.__logger.error("Invalid handshake packet")
            self.__state = ConnectionState.ERROR
            raise MalformedPacket("Invalid handshake packet")

        wallet = self.__wallet

        wallet.token = blake3(
            wallet.epkc.public_bytes(Encoding.Raw, PublicFormat.Raw)
            + wallet.epks.public_bytes(Encoding.Raw, PublicFormat.Raw)
            + wallet.nc
            + wallet.ns
        ).digest()
        client_token = data[8:40]

        self.__logger.debug(f"token: {client_token.hex()}")

        if client_token != wallet.token:
            self.__logger.debug(f"ours : {self.__wallet.token.hex()}")
            self.__logger.error("token mismatch!")
            raise HandsakeError("Token mismatch")

        self.__logger.debug("token: OK")

        self.__key_exchange()

        buffer = self.__verify_and_decrypt(data)
        if buffer is None:
            self.__state = ConnectionState.ERROR
            raise HandsakeError("Invalid handshake packet (missing protocol)")

        messages = self.__reform_messages(buffer)

        if not messages:
            self.__state = ConnectionState.ERROR
            raise HandsakeError("Invalid handshake packet (protocol)")
        self.__state = ConnectionState.CONNECTED
        self.__protocol = messages[0]
        self.__logger.info("Handshake complete")
        self.__logger.debug(f"protocol: {self.__protocol.decode('utf-8')}")

    def __connected(self, data):
        buffer = self.__verify_and_decrypt(data)

        if buffer is not None:
            messages = self.__reform_messages(buffer)
            # self.__logger.debug(f"messages: {messages}")
            # send to all handlers
            for msg in messages:
                asyncio.gather(
                    *(
                        handler(self.__addr, 0, msg)
                        for handler in self.__message_handlers
                    )
                )
                self.__logger.debug(f">>> {trail_off(msg.decode('utf-8'))}")

    def __disconnected(self, data):
        raise NotImplementedError

    def __error(self, data):
        raise NotImplementedError

    def change_max_packets(self, max_size: int):
        self.__incoming_buffer = collections.deque(
            self.__incoming_buffer, maxlen=max_size
        )
        self.__incoming_keys = collections.deque(self.__incoming_keys, maxlen=max_size)

    def handle(self, data: bytes):
        """
        Handles incoming packets.
        :param data: packet data
        """
        self.__last_seen = now()

        if PYTHON_VERSION[1] >= 10:
            match self.__state:
                case ConnectionState.INITIAL:
                    self.__initial(data)
                case ConnectionState.HANDSHAKE:
                    self.__handshake(data)
                case ConnectionState.CONNECTED:
                    self.__connected(data)
                case ConnectionState.DISCONNECTED:
                    self.__disconnected(data)
                case ConnectionState.ERROR:
                    self.__error(data)
        else:
            if self.__state == ConnectionState.INITIAL:
                self.__initial(data)
            elif self.__state == ConnectionState.HANDSHAKE:
                self.__handshake(data)
            elif self.__state == ConnectionState.CONNECTED:
                self.__connected(data)
            elif self.__state == ConnectionState.DISCONNECTED:
                self.__disconnected(data)
            elif self.__state == ConnectionState.ERROR:
                self.__error(data)

    def __flush(self):
        """
        Flushes the outgoing buffer, sending all pending messages.
        Sends multiple packets if necessary. Sends an empty packet if there is no data to send.
        """
        packets = self.__encrypt_and_tag(self.__outgoing_buffer or b"\x00")
        self.__logger.debug(
            f"Sending {len(self.__outgoing_buffer)} bytes in {len(packets)} packets"
        )
        for packet in packets:
            self.__transport.sendto(packet, self.__addr)
        self.__outgoing_buffer.clear()

    def send(self, data: bytes):
        """
        Schedule a message to be sent to the client.
        :param data: data to send
        """
        if self.__state not in (ConnectionState.CONNECTED, ConnectionState.HANDSHAKE):
            return
        self.__logger.debug(f"<<< {trail_off(data.decode('utf-8'))}")
        self.__outgoing_buffer.extend(data)

    def __send_now(self, data: bytes):
        self.__outgoing_buffer.extend(data)
        self.__flush()

    def add_message_handler(self, handler: MessageHandler):
        """
        Adds a message handler. This handler will be called when a message is received.
        :param handler: Awaitable handler function
        """
        self.__message_handlers.add(handler)

    def disconnect(self):
        """
        Disconnects the client.
        """
        self.__state = ConnectionState.DISCONNECTED
        self.__logger.info("Disconnected")
        self.__send_now(b"\x01")
        self.__send_loop_task.cancel()
